/* config.h.  Generated from config.h.in by configure.  */
#define HAVE_MKDTEMP 1
#define HAVE_ON_EXIT 1
/* #undef LIBLDAP21 */
#define LIBLDAP22 1
#define HAVE_OPENSSL 1
/* #undef HAVE_GNUTLS */
#define HAVE_SHA1 1
/* #undef RAND_PSEUDO_BYTES */
#define HAVE_SASL 1
